import { Header } from "../header";
import { HeaderMap } from "../header_map";
import { Param } from "../param";
import { ParameterMap } from "../parameter_map";
import { APIResponse } from "../controllers/api_response";
import { Converter } from "../../utils/util/converter";
/**
* This class is to process the API request and its response.
* Construct the objects that are to be sent as parameters or in the request body with the API.
* The Request parameter, header and body objects are constructed here.
* Process the response JSON and converts it to relevant objects in the library.
*/
declare class CommonAPIHandler {
    private apiPath;
    private param;
    private headers;
    private request;
    private httpMethod;
    private moduleAPIName;
    private contentType;
    private categoryMethod;
    private mandatoryChecker;
    /**
     * This is a setter method to set an API request content type.
     * @param {string} contentType - A String containing the API request content type.
     */
    setContentType(contentType: string): void;
    /**
     * This is a setter method to set the API request URL.
     * @param {string} apiPath - A String containing the API request URL.
     */
    setAPIPath(apiPath: string): void;
    /**
     * This method is to add an API request parameter.
     * @param {Param} paramInstance - A Param instance containing the API request parameter.
     * @param {object} paramValue - An object containing the API request parameter value.
     * @throws {SDKException}
     */
    addParam<T>(paramInstance: Param<T>, paramValue: T): Promise<void>;
    /**
     * This method is to add an API request header.
     * @param {Header} headerInstance - A Header instance containing the API request header.
     * @param {object} headerValue - An object containing the API request header value.
     * @throws {SDKException}
     */
    addHeader<T>(headerInstance: Header<T>, headerValue: T): Promise<void>;
    /**
     * This is a setter method to set the API request parameter map.
     * @param {ParameterMap} param - A ParameterMap class instance containing the API request parameter.
     */
    setParam(param: ParameterMap | undefined | null): void;
    /**
     * This is a getter method to get the Zoho CRM module API name.
     * @returns A String representing the Zoho CRM module API name.
     */
    getModuleAPIName(): string | null;
    /**
     * This is a setter method to set the Zoho CRM module API name.
     * @param {string} moduleAPIName - A String containing the Zoho CRM module API name.
     */
    setModuleAPIName(moduleAPIName: string | null): void;
    /**
     * This is a setter method to set the API request header map.
     * @param {HeaderMap} header - A HeaderMap class instance containing the API request header.
     */
    setHeader(header: HeaderMap | undefined | null): void;
    /**
     * This is a setter method to set the API request body object.
     * @param {object} request - An Object containing the API request body object.
     */
    setRequest(request: any): void;
    /**
     * This is a setter method to set the HTTP API request method.
     * @param {string} httpMethod - A String containing the HTTP API request method.
     */
    setHttpMethod(httpMethod: string): void;
    /**
     * This method is used in constructing API request and response details. To make the Zoho CRM API calls.
     * @param {string} className - A Class containing the method return type.
     * @param {string} encodeType - A String containing the expected API response content type.
     * @see APIHTTPConnector
     * @returns {APIResponse} An instance of APIResponse representing the Zoho CRM API response
     * @throws {SDKException}
     */
    apiCall<T>(className: string, encodeType: string): Promise<APIResponse<T>>;
    private getHeaders;
    /**
     * This method is used to get a Converter class instance.
     * @param {string} encodeType - A String containing the API response content type.
     * @returns A Converter class instance.
     */
    getConverterClassInstance(encodeType: string): Converter | null;
    private setAPIUrl;
    /**
     * This is a getter method to get mandatoryChecker
     * @returns {Boolean} - A Boolean value representing mandatoryChecker
     */
    isMandatoryChecker(): boolean | null;
    /**
     * This is a setter method to set mandatoryChecker
     * @param {Bool} mandatoryChecker - A Boolean value
     */
    setMandatoryChecker(mandatoryChecker: boolean | null): void;
    /**
     * This is a getter method to get the HTTP API request method.
     * @returns {string} A String containing the HTTP API request method.
     */
    getHttpMethod(): string;
    /**
     * This is a getter method to get categoryMethod
     * @returns {String} - A String value representing categoryMethod
     */
    getCategoryMethod(): string;
    /**
     * This is a setter method to set categoryMethod
     * @param {String} categoryMethod - A String value representing categoryMethod
     */
    setCategoryMethod(categoryMethod: string): void;
    /**
     * This is a getter method to get the API request URL.
     * @returns {String} A String containing the API request URL.
     */
    getAPIPath(): string;
}
export { CommonAPIHandler as CommonAPIHandler, CommonAPIHandler as MasterModel };
